//---------------------------------------------------------------------------

#include <vcl.h>
#pragma hdrstop

#include <VclTee.TeeGDIPlus.hpp>

#include "locus.cxx"

#pragma pack(push, 1)
struct ColorPacket
{
	UINT16 	flags;
	UINT8 	stat;
	UINT8 	ambCfg;
	UINT16	clear;
	UINT16	red;
	UINT16	green;
	UINT16	blue;
	UINT16	ir;
	UINT16	irComp;
	UINT16	temp;
};
#pragma pack(pop)

// NB! X=red Y=green Z=blue
enum {
	xyzRed = 0,
	xyzGreen = 1,
	xyzBlue = 2
};

#include "TFrmMax4400xTest.h"
//---------------------------------------------------------------------------
#pragma package(smart_init)
#pragma link "VCLTee.TeeTools"
#pragma resource "*.dfm"
//---------------------------------------------------------------------------

TFrmMax4400xTest *FrmMax4400xTest;
//---------------------------------------------------------------------------

const double c_div = 1000000;

const double c_cmDef[] = {
	67.17,	 11.06, 7.22,
	24.75, 	 71.32,	-17.65,
	-0.0864, -6.37, 79.24
};
//---------------------------------------------------------------------------

const double c_XYZ2eCIERGB[] = {
 2.3706743, -0.9000405, -0.4706338,
-0.5138850,  1.4253036,  0.0885814,
 0.0052982, -0.0146949,  1.0093968
};
//---------------------------------------------------------------------------

const double c_XYZ2sRGB[] = {
 3.1338561, -1.6168667, -0.4906146,
-0.9787684,  1.9161415,  0.0334540,
 0.0719453, -0.2289914,  1.4052427
 };
//---------------------------------------------------------------------------

__fastcall TFrmMax4400xTest::TFrmMax4400xTest(TComponent* Owner) :
TForm(Owner),
m_invalidFill(true),
m_irComp(false),
m_att(1.),
m_gdip(0),
m_assymtotic(0),
m_locus(0),
m_clrt(0),
m_wlTip(0),
m_pnt(0)
{
	memset(m_rgb, 0, sizeof(m_rgb));
	memset(m_rgbOrg, 0, sizeof(m_rgbOrg));
	memset(m_cict, 0, sizeof(m_cict));

	chartInit();
	colorSpaceChartInit();
	// load default correlation matrix
	cmDefaultInit();
	cmToCmEdit();
}
//---------------------------------------------------------------------------

void TFrmMax4400xTest::chartInit()
{
	m_gdip = new TTeeGDIPlus(chart_);
	m_gdip->TeePanel = chart_;
 //	m_gdip->Active = true;

	m_assymtotic = new TLineSeries(chart_);
	m_assymtotic->BeginUpdate();
	m_assymtotic->ParentChart = chart_;
	m_assymtotic->AddXY(0,1);
	m_assymtotic->AddXY(1,0);
	m_assymtotic->Pen->Width = 2;
	m_assymtotic->Color = clGray;
	m_assymtotic->Pen->Style = TPenStyle::psDash;
	m_assymtotic->EndUpdate();

	m_locus = new TLineSeries(chart_);
	m_locus->BeginUpdate();
	m_locus->ParentChart = chart_;
	m_locus->DrawStyle = dsCurve;
	m_locus->OnGetMarkText = locus_GetMarkText;
	m_locus->Pen->Width = 2;
	m_locus->Pen->Color = clGray;
	m_locus->XValues->Order = TChartListOrder::loNone;
	m_locus->YValues->Order = TChartListOrder::loNone;
	m_locus->Pointer->Pen->Color = clGray;
	m_locus->Pointer->Pen->Width = m_locus->Pen->Width;
	m_locus->Pointer->Brush->Style = bsClear;
	m_locus->Pointer->Style = psCircle;
	m_locus->Pointer->Size = 3;
	m_locus->Pointer->Visible = true;
	m_locus->EndUpdate();

	m_clrt = new TLineSeries(chart_);
	m_clrt->BeginUpdate();
	m_clrt->ParentChart = chart_;
	m_clrt->DrawStyle = dsCurve;
	m_clrt->Pen->Width = 2;
	m_clrt->Pen->Color = clNavy;
 	m_clrt->EndUpdate();

	m_wlTip = new TNearestTool(chart_);
	chart_->Tools->Add(m_wlTip);
	m_wlTip->Series = m_locus;
	m_wlTip->Size = 4;
	m_wlTip->Pen->Style = psSolid;
	m_wlTip->Pen->Color = clRed;
	m_wlTip->Pen->Width = 1;
	m_wlTip->Hint = true;
	m_wlTip->OnGetHint = locus_GetTipText;

	m_pnt = new TPointSeries(chart_);
	m_pnt->BeginUpdate();
	m_pnt->ParentChart = chart_;
	m_pnt->Color = clNone;
	m_pnt->Pointer->Style = psCircle;
	m_pnt->Pointer->Size = 4;
	m_pnt->Pointer->Pen->Color = clGray;
	m_pnt->Pointer->Pen->Width = 1;
	m_pnt->AddXY(0,0);
	m_pnt->Shadow->Visible = true;
	m_pnt->Visible = false;
	m_pnt->EndUpdate();
}
//---------------------------------------------------------------------------

TColor TFrmMax4400xTest::rgbToRGB(const double* rgb)
{
	int r = rgb[0]*765.;
	if( r > 255 )
		r = 255;
	else if( r < 0 )
		r = 0;

	int g = rgb[1]*765.;
	if( g > 255 )
		g = 255;
	else if( g < 0 )
		g = 0;

	int b = rgb[2]*765.;
	if( b > 255 )
		b = 255;
	else if( b < 0 )
		b = 0;

	return RGBA(r, g, b, 0);
}
//---------------------------------------------------------------------------

void TFrmMax4400xTest::colorSpaceFillInit()
{
	if( m_invalidFill )
	{
		m_polyRect.init(
			m_locus->CalcXPosValue( m_locus->MinXValue() ),
			m_locus->CalcYPosValue( m_locus->MaxYValue() ),
			m_locus->CalcXPosValue( m_locus->MaxXValue() ),
			m_locus->CalcYPosValue( m_locus->MinYValue() )
		);

		for( int idx = 0; idx < m_locus->Count(); ++idx )
		{
			const TPoint& pnt = TPoint(
				m_locus->CalcXPos(idx)-m_polyRect.Left,
				m_locus->CalcYPos(idx)-m_polyRect.Top
			);
			m_poly.push_back(pnt);
		}
		m_poly.push_back( m_poly[0] );

		m_bmp.reset(new TBitmap);
		m_bmp->PixelFormat = pf32bit;
		m_bmp->HandleType = bmDIB;
		m_bmp->SetSize(m_polyRect.Width(), m_polyRect.Height());
		m_bmp->Transparent = true;
		m_bmp->TransparentColor = clBlack;

		double xyz[3];
		double rgb[3];

		TCanvas* cnv = m_bmp->Canvas;
		cnv->Brush->Color = clBlack;
		cnv->FillRect( TRect(0, 0, m_polyRect.Width(), m_polyRect.Height()) );

		HRGN cr = CreatePolygonRgn(&m_poly[0], m_poly.size(), ALTERNATE);
		SelectClipRgn(cnv->Handle, cr);
		for( int line = 0; line < m_bmp->Height; ++line )
		{
			int pstart = 0;
			int pend = m_bmp->Width;
			xyz[1] = m_locus->YScreenToValue(m_polyRect.Top+line);
			for(int pidx = pstart; pidx < pend; ++pidx)
			{
				xyz[0] = m_locus->XScreenToValue(m_polyRect.Left+pidx);
				xyz[2] = 1. - xyz[0] - xyz[1];

				clrSpaceTransform(c_XYZ2eCIERGB, xyz, rgb);
				cnv->Pixels[pidx][line] = rgbToRGB(rgb);
			}
		}
		SelectClipRgn(cnv->Handle, 0);
		DeleteObject(cr);
		m_invalidFill = false;
	}
}
//---------------------------------------------------------------------------

void TFrmMax4400xTest::colorSpaceChartInit()
{
	m_locus->BeginUpdate();
	m_locus->Clear();

	size_t idx = 0;
	int pidx;
	double rgb[3];
	double xyz[3];
	while( idx < sizeof(sc_locus)/sizeof(sc_locus[0]) )
	{
		++idx; // skip wavelength
		xyz[0] = sc_locus[idx++];
		xyz[1] = sc_locus[idx++];
		xyz[2] = 1. - xyz[0] - xyz[1];
		pidx = m_locus->AddXY(xyz[0], xyz[1]);
		clrSpaceTransform(c_XYZ2eCIERGB, xyz, rgb);
		m_locus->ValueColor[pidx] = rgbToRGB(rgb);
	}
	m_poly.clear();
	m_poly.reserve(m_locus->Count());
	m_locus->EndUpdate();

	m_clrt->BeginUpdate();
	m_clrt->Clear();

	m_clrt->EndUpdate();
}
//---------------------------------------------------------------------------

void TFrmMax4400xTest::cmDefaultInit()
{
	memcpy(m_cm, c_cmDef, sizeof(c_cmDef));
}
//---------------------------------------------------------------------------

void TFrmMax4400xTest::cmToCmEdit()
{
	cm00_->Text = m_cm[0];
	cm10_->Text = m_cm[1];
	cm20_->Text = m_cm[2];
	cm01_->Text = m_cm[3];
	cm11_->Text = m_cm[4];
	cm21_->Text = m_cm[5];
	cm02_->Text = m_cm[6];
	cm12_->Text = m_cm[7];
	cm22_->Text = m_cm[8];
}
//---------------------------------------------------------------------------

void TFrmMax4400xTest::cmEditToCm()
{
	m_cm[0] = cm00_->Text.ToDouble();
	m_cm[1] = cm10_->Text.ToDouble();
	m_cm[2] = cm20_->Text.ToDouble();
	m_cm[3] = cm01_->Text.ToDouble();
	m_cm[4] = cm11_->Text.ToDouble();
	m_cm[5] = cm21_->Text.ToDouble();
	m_cm[6] = cm02_->Text.ToDouble();
	m_cm[7] = cm12_->Text.ToDouble();
	m_cm[8] = cm22_->Text.ToDouble();
}
//---------------------------------------------------------------------------

void TFrmMax4400xTest::clrSpaceTransform(const double* mx, const double* sFrom, double* sTo, double att /*= 1*/)
{
	sTo[0] = (mx[0]*sFrom[0] + mx[1]*sFrom[1] + mx[2]*sFrom[2])*att;
	sTo[1] = (mx[3]*sFrom[0] + mx[4]*sFrom[1] + mx[5]*sFrom[2])*att;
	sTo[2] = (mx[6]*sFrom[0] + mx[7]*sFrom[1] + mx[8]*sFrom[2])*att;
}
//---------------------------------------------------------------------------

void TFrmMax4400xTest::rgb2XYZ()
{
	clrSpaceTransform(m_cm, m_rgb, m_XYZ, m_att/c_div);

	X_->Caption = m_XYZ[0];
	Y_->Caption = m_XYZ[1];
	Z_->Caption = m_XYZ[2];
}
//---------------------------------------------------------------------------

void TFrmMax4400xTest::XYZ2xyz(double X, double Y, double Z, double& x, double& y, double& z)
{
	double sum = X+Y+Z;
	x = X/sum;
	y = Y/sum;
	z = Z/sum;
}
//---------------------------------------------------------------------------

void TFrmMax4400xTest::XYZ2xyz()
{
	XYZ2xyz(m_XYZ[0], m_XYZ[1], m_XYZ[2], m_xyz[0], m_xyz[1], m_xyz[2]);

	xx_->Caption = m_xyz[0];
	yy_->Caption = m_xyz[1];

	m_pnt->BeginUpdate();
	m_pnt->XValue[0] = m_xyz[0];
	m_pnt->YValue[0] = m_xyz[1];
	m_pnt->EndUpdate();

	color_->Invalidate();
}
//---------------------------------------------------------------------------

void TFrmMax4400xTest::colorDataUpdate()
{
	clear_->Caption = m_cict[0];
	r_->Caption = m_rgb[0];
	g_->Caption = m_rgb[1];
	b_->Caption = m_rgb[2];
	ir_->Caption = m_cict[1];
	irc_->Caption = m_cict[2];
	temp_->Caption = m_cict[3];

	rgb2XYZ();
	XYZ2xyz();
}
//---------------------------------------------------------------------------

void TFrmMax4400xTest::attSet(int att)
{
	switch( att )
	{
	case 0:
		m_att = 1;
		break;
	case 1:
		m_att = 4;
		break;
	case 2:
		m_att = 16;
		break;
	case 3:
		m_att = 256;
		break;
	}

	lblAtt_->Caption = m_att;
}
//---------------------------------------------------------------------------

void __fastcall TFrmMax4400xTest::cportPacket_Packet(TObject *Sender, const UnicodeString Str)
{
	static UINT8 s_buff[18];
	for(size_t idx = 0; idx < sizeof(s_buff); ++idx)
		s_buff[idx] =	Str[idx+1];

	ColorPacket* packet = reinterpret_cast<ColorPacket*>(s_buff);

	UINT16 tmp = packet->stat;
	ambint_->Checked = tmp & 0x01;
	pwron_->Checked = tmp & 0x04;
	shdn_->Checked = tmp & 0x08;
	reset_->Checked = tmp & 0x10;

	tmp = packet->flags;
	ESTAT_->Checked = tmp & 0x0001;
	ECFG_->Checked = tmp & 0x0002;
	EAMBCFG_->Checked = tmp & 0x0004;
	ETHRESH_->Checked = tmp & 0x0008;
	EDATAC_->Checked = tmp & 0x0010;
	EDATAR_->Checked = tmp & 0x0020;
	EDATAG_->Checked = tmp & 0x0040;
	EDATAB_->Checked = tmp & 0x0080;
	EIR_->Checked = tmp & 0x0100;
	EIRC_->Checked = tmp & 0x0200;
	ETEMP_->Checked = tmp & 0x0400;

	attSet(static_cast<long>(packet->ambCfg & 0x03));

	m_rgb[0] = packet->red;
	m_rgb[1] = packet->green;
	m_rgb[2] = packet->blue;

	m_cict[0] = packet->clear;
	m_cict[1] = packet->ir;
	m_cict[2] = packet->irComp;
	m_cict[3] = packet->temp;

	doIrComp();

	colorDataUpdate();
}
//---------------------------------------------------------------------------

void TFrmMax4400xTest::irCompSet(bool irComp)
{
	if( m_irComp != irComp )
	{
		m_irComp = irComp;
		doIrComp();
		colorDataUpdate();
	}
}
//---------------------------------------------------------------------------

void TFrmMax4400xTest::doIrComp()
{
	if( m_irComp )
	{
		m_rgb[0] = m_rgbOrg[0] - m_cict[2];
		m_rgb[1] = m_rgbOrg[1] - m_cict[2];
		m_rgb[2] = m_rgbOrg[2];
	}
	else
		memcpy(m_rgbOrg, m_rgb, sizeof(m_rgb));
}
//---------------------------------------------------------------------------

void __fastcall TFrmMax4400xTest::actConfigure_Execute(TObject *Sender)
{
//	cport_->ShowSetupDialog();
}
//---------------------------------------------------------------------------

void __fastcall TFrmMax4400xTest::actConfigure_Update(TObject *Sender)
{
//	actConfigure_->Enabled =  !cport_->Connected;
}
//---------------------------------------------------------------------------

void __fastcall TFrmMax4400xTest::actStartStop_Execute(TObject *Sender)
{
//	cport_->Connected = !cport_->Connected;
}
//---------------------------------------------------------------------------

void __fastcall TFrmMax4400xTest::actStartStop_Update(TObject *Sender)
{
//	actStartStop_->Caption = cport_->Connected ? _T("Stop") : _T("Start");
}
//---------------------------------------------------------------------------

void __fastcall TFrmMax4400xTest::actApplyCm_Execute(TObject *Sender)
{
	cmEditToCm();
}
//---------------------------------------------------------------------------

void __fastcall TFrmMax4400xTest::actApplyCm_Update(TObject *Sender)
{
	actApplyCm_->Enabled =
		cm00_->Modified ||
		cm10_->Modified ||
		cm20_->Modified ||
		cm01_->Modified ||
		cm11_->Modified ||
		cm21_->Modified ||
		cm02_->Modified ||
		cm12_->Modified ||
		cm22_->Modified;
}
//---------------------------------------------------------------------------

void __fastcall TFrmMax4400xTest::locus_GetMarkText(TChartSeries *Sender, int ValueIndex, UnicodeString &MarkText)
{
	MarkText = sc_locus[ValueIndex*3];
	MarkText += _T(" nm");
}
//---------------------------------------------------------------------------

void __fastcall TFrmMax4400xTest::locus_GetTipText(TMarksTipTool *Sender, UnicodeString &MarkText)
{
	MarkText = sc_locus[m_wlTip->Point*3];
	MarkText += _T(" nm");
}
//---------------------------------------------------------------------------

void __fastcall TFrmMax4400xTest::color_Paint(TObject *Sender)
{
	TCanvas* cnv = color_->Canvas;
	cnv->Lock();

	double rgb[3];
	clrSpaceTransform(c_XYZ2sRGB, m_xyz, rgb);

	TColor cl = rgbToRGB(rgb);
	cnv->Brush->Color = cl;

	TRect area(0,0,color_->Width, color_->Height);
	cnv->FillRect(area);
	cnv->Unlock();
}
//---------------------------------------------------------------------------

void __fastcall TFrmMax4400xTest::cport_AfterOpen(TObject *Sender)
{
	m_pnt->Visible = true;
	color_->Visible = true;
}
//---------------------------------------------------------------------------

void __fastcall TFrmMax4400xTest::cport_AfterClose(TObject *Sender)
{
	m_pnt->Visible = false;
	color_->Visible = false;
}
//---------------------------------------------------------------------------

void __fastcall TFrmMax4400xTest::FormClose(TObject *Sender, TCloseAction &Action)
{
//	cport_->Connected = false;
}
//---------------------------------------------------------------------------

void __fastcall TFrmMax4400xTest::actCompIR_Execute(TObject *Sender)
{
	irCompSet(!m_irComp);
}
//---------------------------------------------------------------------------

void __fastcall TFrmMax4400xTest::actCompIR_Update(TObject *Sender)
{
	actCompIR_->Checked = m_irComp;
}
//---------------------------------------------------------------------------

void __fastcall TFrmMax4400xTest::actResetCm_Execute(TObject *Sender)
{
	cmDefaultInit();
	cmToCmEdit();
}
//---------------------------------------------------------------------------

void __fastcall TFrmMax4400xTest::chart_BeforeDrawSeries(TObject *Sender)
{
	colorSpaceFillInit();
	chart_->Canvas->Draw(m_polyRect.Left, m_polyRect.Top, m_bmp.get());
}
//---------------------------------------------------------------------------

void __fastcall TFrmMax4400xTest::actGDIp_Execute(TObject *Sender)
{
	m_gdip->Active = !m_gdip->Active;
}
//---------------------------------------------------------------------------

void __fastcall TFrmMax4400xTest::actGDIp_Update(TObject *Sender)
{
	actGDIp_->Checked = m_gdip->Active;
}
//---------------------------------------------------------------------------

